# -*- coding: utf-8 -*-

# mapWOC
#
# Description: JSON De/Encoders.
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import json

from hc import HC
from hccontrol import HCControl
from message import Message
from redirectorvisit import RedirectorVisit
from scanconfiguration import VisitorConfiguration, VisitorBundle, TransportableSC
from trivialresponse import TrivialResponse
from url import URL, URLList

# TODO Remove usage and definition of 2nd entry in tuples.
"""
Shortcuts to create mapping to/from json.
"""
meta_map_map = [
               (Message, 'Message', ['level', 'text']),
               (URL, 'URL', ['url']),
               (RedirectorVisit, 'RedirectorVisit', ['date', 'url',
                                                      'hcsid', 'peer']),
               (VisitorConfiguration, 'VisitorConfiguration', ['active_window',
                                                               'visitors',
                                                               'num_urls']),
               (VisitorBundle, 'VisitorBundle', ['hcsid', 'number', 'node']),
               (TrivialResponse, 'TrivialResponse', ['error', 'message']),
               (TransportableSC, 'TransportableSC', ['name',
                                                     'comment',
                                                     'urllistIDs',
                                                     'visitor_configuration',
                                                     'description'])
               ]

class MapWOCJSONEncoder(json.JSONEncoder):
    """ Encode MapWOC objects to the JSON format """
    def default(self, python_object):
        """Overridden default encoding."""

        # Use fancy, but probably slow, meta-fu.
        obj = python_object
        for (mclass,name,props) in meta_map_map:
            if isinstance(obj,mclass):
                obj_dict = {'__class__' : 'mapWOC_%s' % name}
                for prop in props:
                    obj_dict[prop] = obj.__getattribute__(prop)
                return obj_dict
        # Else, use handcraft.
        if isinstance(python_object, HCControl) or isinstance(python_object, HC):
            # TODO distinguish between HC/HCControls
            return {
                    '__class__' : 'mapWOC_HC',
                    'hcsid' : python_object.hcsid,
                    'name' : python_object.name,
                    'description' : python_object.description,
                    'active_browser' : python_object.active_browser,
                    'on_nodes' : ",".join(python_object.on_nodes)
                   }
        elif isinstance(python_object, URLList):
            assert("URLList is not meant for transport")
        elif isinstance(python_object, ScanConfiguration):
            assert("ScanConfiguration is not meant for transport")
        else:
            return json.JSONEncoder.default(self, python_object)

def mapwoc_from_json_hook(json_object):
    """ Object Hook to construct mapWOC objects from JSON """

    objClass = json_object.get('__class__')
    if not objClass: return json_object

    for (mclass,name,props) in meta_map_map:
        if objClass == 'mapWOC_%s' % name:
            obj = mclass()
            for prop in props:
                obj.__setattr__(prop, json_object[prop])
            return obj

    # Then, do it by hand.
    if objClass == 'mapWOC_HC':
        hc = HC()
        hc.name = json_object['name']
        hc.hcsid = json_object['hcsid']
        hc.description = json_object['description']
        hc.active_browser = json_object['active_browser']
        hc.on_nodes = json_object['on_nodes'].split(",")
        return hc
    else:
        return json_object

def mapwoc_dejson(string):
    """Convinience shortcut to decode a json string."""
    return json.loads(string, object_hook=mapwoc_from_json_hook)
