#!/usr/bin/python
# -*- coding: utf-8 -*-

# mapWOC
#
# Description: A MasterCtrl with twisted.
#
# Copyright:
# Copyright (C) 2011 Bundesamt für Sicherheit in der Informationstechnik
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 2,
# or, at your option, any later version as published by the Free
# Software Foundation
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

from optparse import OptionParser
from ConfigParser import RawConfigParser, NoSectionError
from time import sleep
import json
import logging

from twisted.internet.protocol import Factory, Protocol
from twisted.protocols.basic import LineReceiver
from twisted.internet import reactor, task

from mapwoc.lib.protocolutil import json_trsuccess, json_trerror
from mapwoc.lib.mapwoc_json import MapWOCJSONEncoder, mapwoc_from_json_hook, mapwoc_dejson

from mapwoc.masterctrl.master import Master

class Requests:
    """ Request Type 'enum'."""
    NODESTATES = 1001
    STATE = 1002
    MESSAGES = 1003
    HCS = 1005
    HC = 1006
    START_SCAN = 1007
    STOP_SCAN = 1008
    VERSION = 1010
    CHECK_SCAN = 1011

cmd_request_map = {'state': Requests.STATE,
                   'nodestates': Requests.NODESTATES,
                   'messages':Requests.MESSAGES,
                   'hcs':Requests.HCS,
                   'hc':Requests.HC,
                   'startscan':Requests.START_SCAN,
                   'stopscan':Requests.STOP_SCAN,
                   'version':Requests.VERSION,
                   'checkscan':Requests.CHECK_SCAN
                   }

def which_cmd(text):
    """Returns the Request field or None."""
    
    simple_cmd = cmd_request_map.get(text)

    if simple_cmd is None:
        # Try to parse more complicated command
        tokens = text.split(" ")
        if len(tokens) == 2 and tokens[0] == 'hc':
            return Requests.HC,[tokens[1]]
        else:
            return None,[]
    else:
        return simple_cmd,[]

# TODO document protocol.
class MasterRequest(LineReceiver):

    def connectionMade(self):
        """Inits empty buffer and command."""

        self.read_data = ''
        self.cmd = None

    def lineReceived(self, line):
        """Reacts on a received line."""

        self.read_data += line

        if not self.cmd:
            self.cmd,args =  which_cmd(self.read_data.strip())
            if self.cmd is Requests.NODESTATES:
                answer = self.factory.master.query_nodes_state()
                self.__answer_json(answer)
            elif self. cmd is Requests.STATE:
                answer = self.factory.master.state() 
                self.__answer_json(answer)
            elif self.cmd is Requests.MESSAGES:
                answer = self.factory.master.get_last_msgs(2)
                self.__answer_json(answer)
            elif self.cmd is Requests.HCS:
                answer = self.factory.master.get_hcsid_on_nodes()
                self.__answer_json(answer)
            elif self.cmd is Requests.HC:
                answer = self.factory.master.get_hc(args[0])
                self.__answer_json(answer)
            elif self.cmd is Requests.START_SCAN:
                # Read the scan configuration object.
                self.read_data = ''
            elif self.cmd is Requests.STOP_SCAN:
                self.factory.master.stop_scan()
                self.__answer(json_trsuccess())
            elif self.cmd is Requests.VERSION:
                self.__answer(self.factory.master.VERSION)
            elif self.cmd is Requests.CHECK_SCAN:
                self.read_data = ''
            else:
                print "Not (yet) a command: _%s_" % self.read_data
        else:
            #TODO its dangerous to assume no newlines are in e.g. json

            if self.cmd is Requests.START_SCAN:
                scanconfig = json.loads(self.read_data, object_hook=mapwoc_from_json_hook)
                if self.factory.master.start_scan(scanconfig):
                    self.__answer(json_trsuccess())
                else:
                    self.__answer(json_trerror(333, "Did not start scan"))
            elif self.cmd is Requests.CHECK_SCAN:
                scanconfig = mapwoc_dejson(self.read_data)
                if self.factory.master.check_scan(scanconfig):
                    self.__answer(json_trsuccess())
                else:
                    self.__answer(json_trerror(444, "Nodes can not start scan"))
            

    def __answer(self, answer):
        """Sends answer through the wire and closes connection afterwards."""

        self.transport.write(answer)
        self.transport.loseConnection()


    def __answer_json(self, answer):
        """Sends json encoded answer through the wire and closes the
        connection afterwards.
        """

        json_str = MapWOCJSONEncoder().encode(answer)
        self.__answer(json_str)



class MasterResponseFactory(Factory):
    """Factory producing MasterRequest Protocols."""

    protocol = MasterRequest

    def __init__(self, master):
        self.master = master

def setup_logging(quiet, verbose):
    """Setup logging."""

    mainlogger = logging.getLogger('masterctrl')
    mainlogger.addHandler(logging.FileHandler('master.log'))
    mainlogger.addHandler(logging.StreamHandler())
    masterlogger = logging.getLogger('master')
    masterlogger.addHandler(logging.FileHandler('master.log'))
    masterlogger.addHandler(logging.StreamHandler())
    serverlogger = logging.getLogger('masterctrlsrv')
    serverlogger.addHandler(logging.FileHandler('master.log'))
    serverlogger.addHandler(logging.StreamHandler())

    if verbose:
        mainlogger.setLevel(logging.DEBUG)
        masterlogger.setLevel(logging.DEBUG)
        serverlogger.setLevel(logging.DEBUG)
    elif quiet:
        mainlogger.setLevel(logging.WARNING)
        masterlogger.setLevel(logging.WARNING)
        serverlogger.setLevel(logging.WARNING)
    else:
        mainlogger.setLevel(logging.INFO)
        masterlogger.setLevel(logging.INFO)
        serverlogger.setLevel(logging.INFO)

def main():
    """Startup a Reactor and MasterCtrl."""

    description = ('mapwoc-master has control over one or more mapwoc-nodes.  '
                   'mapwoc-master listens on an adress for commands and '
                   'information requests (e.g. from mapwocgui).')

    parser = OptionParser(
              version='mapwoc-master (twisted wrapper) %s' % Master.VERSION,
              description=description)
    parser.add_option("", "--config", dest="configfile",
                      help="Path to configuration file.",
                      default="/etc/mapwoc/master/master.config")
    parser.add_option("-v", "--verbose", dest="verbose", action="store_true",
                      help="Be verbose and chatty on stdout and logfile.")
    parser.add_option("-q", "--quiet", dest="quiet", action="store_true",
                      help="Be quiet, only print warnings and"
                           " errors to stdout and logfile.")
    (options, args) = parser.parse_args()
    if options.quiet and options.verbose:
        print "--verbose takes precedence over --quiet"
    
    setup_logging(options.quiet, options.verbose)
    
    try:
        master = Master(options.configfile)
    except NoSectionError:
         print "Given configuration file %s is invalid." % options.configfile
         exit(2)

    config = RawConfigParser()
    config.read(options.configfile)
    port = int((config.get('config', 'ctrl_port') or 8558))
    f = MasterResponseFactory(master)
    reactor.listenTCP(port, f)

    master_polls_nodes = task.LoopingCall(master.check_state_polls)
    master_polls_nodes.start(10)

    try:
        reactor.run()
    except:
        #reactor.stop()
        pass

    exit(0)

if __name__ == '__main__':
    main()

